import dateutil.parser
import uuid
from typing import Tuple, Dict, List, Union
from marshmallow.fields import Nested, String, Str, Int, Integer, UUID, DateTime, Boolean, Bool


class Filter(Nested):
    """
    Type for define filterable fields
    """
    _MARSHMALLOW_MAP = {
        String: 'string',
        Str: 'string',
        Integer: 'number',
        Int: 'number',
        UUID: 'uuid',
        DateTime: 'datetime',
        Boolean: 'boolean',
        Bool: 'boolean'
    }

    def __init__(self, *args, **kwargs):
        self._filterable_fields = kwargs.get("filterable")
        self._entity_schema = args[0]
        super().__init__(*args, **kwargs)

    def _serialize(self, nested_obj, attr, obj, **kwargs):
        result = {}
        for field_name, field_value in nested_obj.items():
            if isinstance(field_value, dict):
                result[field_name] = self._serialize(field_value, field_name, obj, **kwargs)
            else:
                result[field_name] = self._process_payload(
                    self._entity_schema().declared_fields[attr].__class__,
                    field_value
                )

        return result

    @staticmethod
    def _process_payload(type, value):
        _map = {
            DateTime: dateutil.parser.parse,
            String: str,
            Str: str,
            Integer: int,
            Int: int,
            UUID: uuid.UUID,
            Boolean: bool,
            Bool: bool
        }
        return _map[type](value)

    def _jsonschema_type_mapping(self):
        properties, required_fields = self._build_schema_attributes()
        return {
            "type": "object",
            "properties": properties,
            "required": required_fields
        }

    def _build_schema_attributes(self) -> Tuple[Dict[str, Union[str, Dict]], List]:
        """Build attributes to build schema for filter field"""

        entity_schema = self._entity_schema

        if callable(entity_schema):
            entity_schema = entity_schema()

        properties = {}
        required_fields = []

        for name, cls in entity_schema.declared_fields.items():
            if name in self._filterable_fields:
                if entity_schema.fields.get(name).required:
                    required_fields.append(name)
                properties[name] = dict({
                    "$ref": "#/definitions/operators/{}".format(self._MARSHMALLOW_MAP.get(cls.__class__)),
                    "description": entity_schema.fields.get(name).metadata.get("description")
                })

        return properties, required_fields
