from uuid import UUID, uuid4
from aiohttp.web_urldispatcher import View
from aiohttp.web_response import Response
from .registry import OperationRegistry, REGISTRY_SETTINGS_KEY
from .exceptions import JsonRpcException, JSONRPC_ERROR
from .validation import SchemaValidator
from .schema.dump import Dumper
from prometheus_client import Histogram
from prometheus_async.aio import time

REQ_TIME = Histogram("req_time_seconds", "time spent in requests")


class JsonRPCView(View):
    @property
    def operations_registry(self) -> OperationRegistry:
        """Auxiliary property to fast access to operation registry"""

        return self.request.app[REGISTRY_SETTINGS_KEY]


class HealthCheckView(View):
    async def get(self) -> Response:
        result = await self.request.app['health_check_function'](self.request.app)
        return Response(body='OK', status=200 if result else 504)


class SelfDescribeView(JsonRPCView):
    async def post(self) -> Response:
        params = await self.request.json()
        method = params.get("method")
        params = params.get("params")
        if method == "specs.operations":
            dumper = Dumper(self.operations_registry)
            if params and params.get("name"):
                return Response(body=dumper.operation(params.get("name")))
            return Response(body=dumper.registry())


class ApiView(JsonRPCView):
    """Main view to handle requests"""

    @time(REQ_TIME)
    async def post(self) -> Response:
        params = await self.request.json()
        request_id = params.get("id")
        payload = params.get("params")

        try:
            operation = self.operations_registry.get(params.get("method"))
            SchemaValidator(self.operations_registry).validate(operation.request_schema, payload)
            handler = operation.handler(app_ctx=self.request.app)
            response = await handler.handle(payload=payload)
        except JsonRpcException as e:
            return self._error_response(request_id, e)
        except Exception as e:
            import traceback
            traceback.print_exc()
            return self._error_response(request_id, JsonRpcException(
                code=JSONRPC_ERROR,
                message="Unhandled JSON RPC error"
            ))

        return self._success_response(request_id, response)

    @property
    def _trace_id(self) -> UUID:
        try:
            trace_id = UUID(self.request.headers.get('X-TRACE-ID'))
        except Exception:
            trace_id = uuid4()

        return trace_id

    @staticmethod
    def _success_response(request_id, data):
        response = {
            "id": request_id,
            "jsonrpc": "2.0",
            "result": data
        }

        return Response(body=response)

    @staticmethod
    def _error_response(request_id, error: JsonRpcException):
        response = {
            "error": {
                "code": error.code,
                "data": error.context,
                "message": error.message
            },
            "id": request_id,
            "jsonrpc": "2.0"
        }
        return Response(body=response)
